/*
 * T-Plan Robot Enterprise, automated testing tool.
 * Copyright (C) 2016 T-Plan Limited (http://www.t-plan.com),
 * All rights reserved.
 */
package com.tplan.extent;

import com.aventstack.extentreports.ExtentReports;
import com.aventstack.extentreports.ExtentTest;
import com.aventstack.extentreports.MediaEntityBuilder;
import com.aventstack.extentreports.Status;
import com.aventstack.extentreports.reporter.ExtentHtmlReporter;
import com.aventstack.extentreports.reporter.KlovReporter;
import com.tplan.robot.ApplicationSupport;
import com.tplan.robot.scripting.ScriptEvent;
import com.tplan.robot.scripting.ScriptListener;
import com.tplan.robot.scripting.ScriptManager;
import com.tplan.robot.scripting.ScriptingContext;
import com.tplan.robot.scripting.commands.CommandEvent;
import com.tplan.robot.scripting.commands.CommandListener;
import com.tplan.robot.scripting.commands.OutputObject;
import com.tplan.robot.scripting.commands.ScriptLogger;
import com.tplan.robot.scripting.commands.impl.ScreenshotCommand;
import com.tplan.robot.scripting.commands.impl.ScriptInfo;
import com.tplan.robot.scripting.commands.impl.StepCommand;
import com.tplan.robot.scripting.commands.impl.StepInfo;
import com.tplan.robot.scripting.commands.impl.WarningCommand;
import com.tplan.robot.util.FileUtils;
import com.tplan.robot.util.Utils;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;

/**
 * Provides the ExtentReport 3.x and Klov report server connectivity.
 */
public class KlovReportImpl implements ExtentReport, ScriptListener, CommandListener, PropertyChangeListener {

    public static final String UNTITLED_TEST_NAME = "Untitled test";

    private File reportFile = null;
    private ExtentHtmlReporter htmlReporter;
    private ExtentReports extentReport;
    private ExtentTest extentTest;
    private int contextNo;
    private boolean recordLogs;
    private boolean serverMode;
    private boolean mimeErrorMsgShown = false;

    public KlovReportImpl() throws ClassNotFoundException {
        Class.forName("com.aventstack.extentreports.reporter.KlovReporter");
    }

    @Override
    public void startReport(ScriptingContext context, String file, String testName, String testDesc, String projectName,
            String mongoHost, int mongoPort, String klovHost, int klovPort, boolean recordLogs) {
        // If the "Script->Disable Execution Output" flag is on skip the report
        if (Boolean.valueOf(System.getProperty(ScriptManager.OUTPUT_DISABLED_FLAG))) {
            return;
        }

        if (reportFile == null) {
            // Save the context sequence number
            contextNo = context.getSeqNo();

            // ExtentX connection (since v0.3)
            serverMode = mongoHost != null && klovHost != null;

            // Save the flag controlling log recording
            this.recordLogs = recordLogs;

            // Get the report file path
            reportFile = new File(file);
            if (!reportFile.isAbsolute()) {
                // Resolve relative path against the script report dir specified 
                // by the _REPORT_DIR variable in the context
                reportFile = new File(context.getVariableAsString(ScriptingContext.IMPLICIT_VARIABLE_REPORT_DIR), file);
            }

            // Set on the append mode if the specified report 
            // file already exists
            boolean append = reportFile.exists();

            // Create the ExtentReports instance
            extentReport = new ExtentReports();

            // Create & attach the HTML reporter
            htmlReporter = new ExtentHtmlReporter(reportFile);
            htmlReporter.setAppendExisting(append);
            extentReport.attachReporter(htmlReporter);

            // Klov connection (since v0.4)
            if (serverMode) {
                KlovReporter klov = new KlovReporter();
                klov.setReportName(testName);
                if (projectName != null) {
                    klov.setProjectName(projectName.trim());
                }
                klov.initMongoDbConnection(mongoHost, mongoPort);
                StringBuilder b = new StringBuilder();
                if (!klovHost.contains("://")) {
                    b.append("http://");
                }
                b.append(klovHost).append(':').append(klovPort);
                String url = b.toString();
                klov.setKlovUrl(url);
                extentReport.attachReporter(klov);
                
                b.insert(0, "The Extent report is being uploaded to the Klov server at <a href=\"");
                b.append("\">").append(url).append("</a>");
                context.getExecutionLog().log(Level.INFO, b.toString());
            }

            // Make sure that the test name and description are 
            // not null (Extent throws an NPE otherwise)
            if (testName == null) {
                testName = getScriptFileName(context);
                if (testName == null) {
                    testName = UNTITLED_TEST_NAME;
                }
            }
            if (testDesc == null) {
                testDesc = "";
            }
            extentTest = extentReport.createTest(testName, testDesc);

            // Save this instance to a list in the context. We need it in 
            // the context because this way it can be accessed by other scripts 
            // in this plugin such as Flush and Set. It must be a list because 
            // one script may theoretically create multiple Extent reports.
            List<ExtentReport> reports = (List<ExtentReport>) context.get(CONTEXT_KEY_EXTENT_REPORT_LIST);
            if (reports == null) {
                reports = new ArrayList<ExtentReport>();
                context.put(CONTEXT_KEY_EXTENT_REPORT_LIST, reports);
            }
            reports.add(this);

            // Set the test start date to the one recorded in the Robot context
            Date date = (Date) context.get(ScriptingContext.CONTEXT_EXECUTION_START_DATE);
            if (date != null) {
                extentTest.getModel().setStartTime(date);
            }

            // Add any reportable objects created so far.
            // These will unfortunately have a wrong time stamp because 
            // the Extent API doesn't support custom time
            List<OutputObject> lo = context.getOutputObjects();
            if (lo != null && !lo.isEmpty()) {
                for (OutputObject oo : lo) {
                    addObjectToExtentReport(oo, context);
                }
            }

            // Register as the CommandListener to receive events on reportable objects 
            // creation (screen shots, steps, scripts, ...). See commandEvent() below.
            context.getScriptManager().addCommandListener(this);

            // Register as the ScriptListener to receive events of the script execution 
            // life cycle (start, pause, stop, finish). See scriptEvent() below
            context.getScriptManager().addScriptListener(this);

            // Register with the execution log manager to receive new logs
            context.getExecutionLog().addHtmlLogListener(this);

            // Add a log mentioning the report location to the script execution log
            StringBuilder msg = new StringBuilder();
            msg.append("Test results are being saved to the Extent report at ");
            msg.append("<a href=\"file://").append(reportFile).append("\">").append(reportFile).append("</a>");
            context.getExecutionLog().logInfo(msg.toString());
        }
    }

    @Override
    public void flushReport() {
        ExtentReports rep = extentReport;
        if (rep != null) {
            rep.flush();
        }
    }

    @Override
    public void addAuthor(String author) {
        ExtentTest test = extentTest;
        if (test != null) {
            test.assignAuthor(author);
        }
    }

    @Override
    public void addSystemInfo(String param, String value) {
        ExtentReports rep = extentReport;
        if (rep != null) {
            rep.setSystemInfo(param, value);
        }
    }

    @Override
    public void addCategory(String category) {
        ExtentTest test = extentTest;
        if (test != null) {
            test.assignCategory(category);
        }
    }

    @Override
    public void setConfiguration(File config) {
        ExtentReports rep = extentReport;
        if (rep != null) {
            //rep.config().statusConfigurator().loadConfig(config);
        }
    }

    /**
     * Fired on the test script status change (start, stop, pause, resume, ...).
     *
     * @param event a script event.
     */
    @Override
    public void scriptEvent(ScriptEvent event) {
        if (event.getContext().getSeqNo() == contextNo) {
            switch (event.getType()) {
                case ScriptEvent.SCRIPT_EXECUTION_FINISHED:  // The script has finished 
                    try {
                        // Add the final log
                        addFinalLog(event.getContext());

                        // Close the current test
                        extentTest.getModel().end();
                        
                        // This is a hack to make the test follow the Robot model 
                        // of pass/fail. The exit code of 0 is PASS and anything else 
                        // is FAIL. Extent on the other hand fails any test containing 
                        // a failed step.
                        Status robotStatus = event.getContext().getExitCode() == 0 ? Status.PASS : Status.FAIL;
                        extentTest.getModel().setStatus(robotStatus);

                        // Write the report to the file and close it
                        extentReport.flush();
                        extentReport.removeTest(extentTest);

                        // Improvement in 0.2 - when running in the CLI mode print out the 
                        // console message like "Report is available at ..." which gets 
                        // printed for the default XML/HTML reports
                        if (event.getScriptManager().isConsoleMode()) {
                            String outParam = System.getProperty(ScriptManager.OUTPUT_DISABLED_FLAG);
                            boolean outputDisabled = outParam != null && outParam.equals("true");
                            if (!outputDisabled) {
                                System.out.println(MessageFormat.format(ApplicationSupport.getString(
                                        "reportProvider.default.cliMsgReportAvailable"), Utils.getFullPath(reportFile)));
                            } else {
                                System.out.println(ApplicationSupport.getString("reportProvider.default.cliMsgReportOff"));
                            }
                        } // -- Improvement end

                    } catch (Throwable t) {
                        t.printStackTrace();
                        event.getContext().getExecutionLog().logSevere("Failed to create the Extent report at "
                                + reportFile + " for " + t.getClass().getName() + ": " + t.getMessage(), false);
                    } finally {
                        // Final cleanup
                        // Remove the listeners 
                        event.getScriptManager().removeCommandListener(this);
                        event.getScriptManager().removeScriptListener(this);
                        event.getContext().getExecutionLog().removeHtmlLogListener(this);

                        // Remove this generator from the list in the context to allow
                        // the Garbage Collector reclaim it
                        List<ExtentReport> reports = (List<ExtentReport>) event.getContext().get(CONTEXT_KEY_EXTENT_REPORT_LIST);
                        if (reports != null) {
                            reports.remove(this);
                        }
                    }
                    break;
            }
        }
    }

    /**
     * Implementation of the {@link CommandListener} interface. It allows us to
     * receive notifications of new screen shots, warnings and test steps
     * created by the executing script.
     *
     * @param e a command event.
     */
    @Override
    public void commandEvent(CommandEvent e) {
        // When a new reportable object is created Robot fires a command 
        // event as follows. The object is referenced in the event.
        if (e.getContext().getSeqNo() == contextNo && e.getActionCode().equals(CommandEvent.OUTPUT_CHANGED_EVENT)) {
            Object o = e.getCustomObject();
            if (o instanceof OutputObject) {
                addObjectToExtentReport((OutputObject) o, e.getContext());
            }
        }
    }

    /**
     * Implementation of the {@link PropertyChangeListener} interface. It allows
     * us to receive logs from the execution log manager.
     *
     * @param evt a property event.
     */
    @Override
    public void propertyChange(PropertyChangeEvent evt) {
        if (recordLogs && ScriptLogger.PROPERTY_CHANGE_NEW_LOG.equals(evt.getPropertyName())) {
            // New execution log added. Unfortunately there's no direct way 
            // allowing us to obtain the script level (log status).
            extentTest.log(Status.INFO, evt.getNewValue().toString());
        }
    }

    /**
     * Add a single reportable object created by a Robot script to the Extent
     * report.
     *
     * @param oo a reportable object.
     */
    private void addObjectToExtentReport(OutputObject oo, ScriptingContext context) {
        StringBuilder b;
        switch (oo.getType()) {
            case OutputObject.TYPE_SCREENSHOT:
                // A screenshot created by the Screenshot command
                // Create the screenshot log (description plus the <img> tag)
                ScreenshotCommand.ScreenshotInfo si = (ScreenshotCommand.ScreenshotInfo) oo;
                b = new StringBuilder();
                b.append(si.description == null ? "" : si.description);

                // Fix in 0.2 - use the relative image path to make it display correctly
                // from an isolated document root managed by a web server or Continuous Integration
                // b.append(extentTest.addScreenCapture(si.imageFile.getAbsolutePath()));
                String imgPath = null;
                IOException err = null;
                try {
                    imgPath = FileUtils.getRelativePath(si.imageFile, reportFile);
                } catch (IOException ioe) {
                    err = ioe;
                }
                // No relative image path -> fall back to the full path and log a warning message 
                if (imgPath == null) {
                    imgPath = si.imageFile.getAbsolutePath();
                    StringBuilder msg = new StringBuilder("Failed to determine relative path of image ");
                    msg.append(si.imageFile).append(" to the report file at ").append(reportFile);
                    if (err != null) {
                        msg.append(" (").append(err.getClass().getName()).append(": ").append(err.getMessage() + ")");
                    }
                    context.getExecutionLog().log(Level.WARNING, msg.toString());
                }
                // This works with the ExtentX HTML file but not with Klov due to their bug.
                // See: Problems attaching screenshots to the log
                // https://github.com/anshooarora/klov/issues/45
                try {
                    extentTest.log(Status.INFO, b.toString(), MediaEntityBuilder.createScreenCaptureFromPath(si.imageFile.getAbsolutePath()).build());
                } catch (NoSuchMethodError ex) {
                    if (!mimeErrorMsgShown) {
                        mimeErrorMsgShown = true;
                        context.getExecutionLog().log(Level.WARNING, "Failed to upload the screen shot to Klov because you didn't put the Extent libraries to the beginning of the "
                                + ApplicationSupport.APPLICATION_NAME + " class path. See the <a href=\"http://www.t-plan.com/robot/docs/plugins/extentreports.html#report\">plugin docs</a> for details.");
                    }
                } catch (Throwable ex) {
                    ex.printStackTrace();
                }
                break;
            case OutputObject.TYPE_WARNING:
                // A warning (rarely used) created by the Warning command
                extentTest.log(Status.WARNING, ((WarningCommand.WarningInfo) oo).description);
                break;
            case OutputObject.TYPE_LOG:
                // Unimplemented in Robot 4.x. Logs are being saved to the execution log
                // directly and this event type gets never fired. We populate logs to the 
                // Extent report from the propertyChanged() method above
                break;
            case StepInfo.TYPE_STEP:
                StepInfo sti = (StepInfo) oo;
                Status status = mapStatus(sti.getResult());
                // TODO: Add the other step attributes such as expected/actual/notes.
                //       As this is not supported by Extent we would have to embed  
                //       it into the text, for example in form of an HTML table.
                extentTest.log(status, sti.getName());
                break;
            case ScriptInfo.TYPE_SCRIPT:  // Not supported -> ignore
                break;
        }
    }

    /**
     * Map the Robot test step status onto the Extent one.
     *
     * @param result the Robot step status (result).
     * @return the corresponding Extent status.
     */
    private Status mapStatus(String result) {
        if (result != null) {
            if (result.equalsIgnoreCase(StepCommand.PARAM_RESULT_FAIL)) {
                return Status.FAIL;
            } else if (result.equalsIgnoreCase(StepCommand.PARAM_RESULT_PASS)) {
                return Status.PASS;
            } else if (result.equalsIgnoreCase(StepCommand.PARAM_RESULT_NOT_AVAILABLE)) {
                return Status.SKIP;
            } else if (result.equalsIgnoreCase(StepCommand.PARAM_RESULT_NOT_TESTED)) {
                return Status.SKIP;
            }
        }
        return Status.ERROR;
    }

    private String getScriptFileName(ScriptingContext context) {
        String scriptFileName = null;
        URI uri = context.getInterpret().getURI();
        if (uri != null) {
            scriptFileName = new File(uri).getName();
        }
        return scriptFileName;
    }

    /**
     * Add the final log with the script exit code and links to other reports to
     * the current Extent report.
     *
     * @param context the script context.
     */
    private void addFinalLog(ScriptingContext context) {
        // Look for additional resources created by the script, 
        // such as the execution log and native Robot reports.
        StringBuilder reports = new StringBuilder();
        File logFile = new File(reportFile.getParent(), ScriptLogger.LOG_FILE_NAME);
        if (logFile.exists()) {
            reports.append("execution log (<a href=\"").append(ScriptLogger.LOG_FILE_NAME).append("\">");
            reports.append(ScriptLogger.LOG_FILE_NAME).append("</a>)");
        }

        // Robot native reports create variables with the report path
        String nativeReport = context.getVariableAsString(ScriptingContext.REPORT_REPORT_FILE);
        if (nativeReport != null) {
            // There's a report -> log it. Versions up to 4.1.2 populated 
            // just the _REPORT_FILE variable. Version 4.1.3 supports multiple reports 
            // and populates additional variables like _REPORT_FILE_1, _REPORT_FILE_2 etc.
            // The code below handles all versions.
            int reportNo = 1;
            String fileName;
            do {
                if (reports.length() > 0) {
                    reports.append(", ");
                }
                fileName = new File(nativeReport).getName();
                reports.append("Robot report (<a href=\"").append(fileName).append("\">");
                reports.append(fileName).append("</a>)");

                // Look for another report by the indexed variable
                nativeReport = context.getVariableAsString(ScriptingContext.REPORT_REPORT_FILE + "_" + (++reportNo));
            } while (nativeReport != null);
        }

        // Log the resources to the Extent report
        if (reports.length() > 0) {
            reports.insert(0, " Other reports: ");
        }

        // Get the script file name
        String scriptFileName = getScriptFileName(context);

        // Construct the closing log
        StringBuilder log = new StringBuilder();
        if (scriptFileName != null) {
            log.append("The <code>").append(scriptFileName).append("</code> test script");
        } else {
            log.append("The test script");
        }
        log.append(" finished with the exit code of ").append(context.getExitCode()).append(".");
        if (reports.length() > 0) {
            log.append(reports);
        }

        // If the Robot's exit code indicates a failure while the Extent 
        // report says "pass" we create a fake failed step to make it fail too.
        Status logStatus = Status.INFO;
        if (context.getExitCode() != 0) {
            logStatus = Status.FAIL;
        }

        extentTest.log(logStatus, log.toString());
    }
}
