/*
 * T-Plan Robot Enterprise, automated testing tool.
 * Copyright (C) 2016 T-Plan Limited (http://www.t-plan.com),
 * All rights reserved.
 */
package com.tplan.extent;

import com.tplan.robot.gui.components.StringComponent;
import com.tplan.robot.preferences.Preference;
import com.tplan.robot.preferences.PreferenceComponentManager;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/**
 * Class testing whether the specified MongoDB address is alive. It is used to
 * deliver functionality of the "Test connection" button displayed in the Report
 * script parameter editor.
 * 
 * @since 0.3
 */
class MongoDBTester implements PreferenceComponentManager, ActionListener, Runnable {

    /**
     * The default Mongo DB port of 27017.
     */
    static final int MONGODB_DEFAULT_PORT = 27017;
    static final int KLOV_DEFAULT_PORT = 80;
    private StringComponent field;
    private String server;
    private int port;
    private JLabel message;

    @Override
    public void componentCreated(Preference p, Object component, Object container) {
        if (component instanceof StringComponent) {
            StringComponent sc = (StringComponent) component;
            sc.getButton().addActionListener(this);
            sc.getButton().setText("Test connection");
            field = sc;
        }
    }

    @Override
    public void goingToLoad(Preference p, Object component, Object container, Object configuration) {
    }

    @Override
    public void savePerformed(Preference p, Object component, Object container, Object configuration) {
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        if (message == null) {
            message = new JLabel();
        }
        String msg = null;
        String text = field.getText();
        server = parseServer(text);

        try {
            port = parsePort(text, MONGODB_DEFAULT_PORT);
        } catch (IllegalArgumentException ex) {
            msg = ex.getMessage();
        }

        if (server == null || server.trim().isEmpty()) {
            msg = "Invalid address: missing host name";
        }
        if (msg == null) {
            msg = "Wait while testing the connection to " + server + ":" + port + "...                               ";
            new Thread(this).start();
        }
        message.setText(msg);

        JOptionPane.showMessageDialog(
                SwingUtilities.getWindowAncestor((Component) e.getSource()),
                message,
                "Testing Connection...",
                JOptionPane.INFORMATION_MESSAGE);
    }

    @Override
    public void run() {
        String msg = "The " + server + ":" + port + " address is alive";
        try {
            Socket sock = new Socket(server, port);
            // OPTIONAL ENHANCEMENT: Test that it's the MongoDB at the specified address
        } catch (UnknownHostException ex) {
            msg = "I/O Error: Unable to determine the host IP address";
        } catch (IOException ex) {
            msg = "I/O Error: Unable to connect to the specified address";
        }
        final String text = msg;
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                message.setText(text);
            }
        });
    }

    static String parseServer(String address) {
        if (address == null) {
            return null;
        }
        int i = address.indexOf(':');
        if (i >= 0) {
            address = address.substring(0, i);
        }
        address = address.trim();
        return address.isEmpty() ? null : address;
    }

    static int parsePort(String address, int defaultPort) throws IllegalArgumentException {
        if (address != null) {
            int i = address.indexOf(':');
            if (i >= 0) {
                try {
                    i = Integer.parseInt(address.substring(i + 1));
                } catch (NumberFormatException e) {
                    throw new IllegalArgumentException("Invalid address: the port is not a valid number");
                }
                if (i < 0 || i > 65535) {
                    throw new IllegalArgumentException("Port number must be between 0 and 65535");
                }
                return i;
            }
        }
        return defaultPort;
    }
}
